import enum

from flask import Flask
from flask_cors import CORS
from datetime import datetime, date
from flask.json import JSONEncoder, jsonify
from importlib_metadata import entry_points
from sqlalchemy.exc import OperationalError
from werkzeug.exceptions import HTTPException

from flaschengeist import logger
from flaschengeist.utils.hook import Hook
from flaschengeist.plugins import AuthPlugin
from flaschengeist.config import config, configure_app


class CustomJSONEncoder(JSONEncoder):
    def default(self, o):
        try:
            # Check if custom model
            return o.serialize()
        except AttributeError:
            pass
        if isinstance(o, datetime) or isinstance(o, date):
            return o.isoformat()
        if isinstance(o, enum.Enum):
            return o.value
        try:
            # Check if iterable
            iterable = iter(o)
        except TypeError:
            pass
        else:
            return list(iterable)
        return JSONEncoder.default(self, o)


@Hook("plugins.loaded")
def load_plugins(app: Flask):
    app.config["FG_PLUGINS"] = {}

    for entry_point in entry_points(group="flaschengeist.plugins"):
        logger.debug(f"Found plugin: {entry_point.name} ({entry_point.dist.version})")

        if entry_point.name == config["FLASCHENGEIST"]["auth"] or (
            entry_point.name in config and config[entry_point.name].get("enabled", False)
        ):
            logger.debug(f"Load plugin {entry_point.name}")
            try:
                plugin = entry_point.load()(entry_point, config=config.get(entry_point.name, {}))
                if hasattr(plugin, "blueprint") and plugin.blueprint is not None:
                    app.register_blueprint(plugin.blueprint)
            except:
                logger.error(
                    f"Plugin {entry_point.name} was enabled, but could not be loaded due to an error.",
                    exc_info=True,
                )
                continue
            if isinstance(plugin, AuthPlugin):
                if entry_point.name != config["FLASCHENGEIST"]["auth"]:
                    logger.debug(f"Unload not configured AuthPlugin {entry_point.name}")
                    del plugin
                    continue
                else:
                    logger.info(f"Using authentication plugin: {entry_point.name}")
                    app.config["FG_AUTH_BACKEND"] = plugin
            else:
                logger.info(f"Using plugin: {entry_point.name}")
            app.config["FG_PLUGINS"][entry_point.name] = plugin
        else:
            logger.debug(f"Skip disabled plugin {entry_point.name}")
    if "FG_AUTH_BACKEND" not in app.config:
        logger.fatal("No authentication plugin configured or authentication plugin not found")
        raise RuntimeError("No authentication plugin configured or authentication plugin not found")


def create_app(test_config=None, cli=False):
    app = Flask("flaschengeist")
    app.json_encoder = CustomJSONEncoder
    CORS(app)

    with app.app_context():
        from flaschengeist.database import db, migrate

        configure_app(app, test_config)
        db.init_app(app)
        migrate.init_app(app, db, compare_type=True)
        load_plugins(app)

    @app.route("/", methods=["GET"])
    def __get_state():
        from . import __version__ as version

        return jsonify({"plugins": app.config["FG_PLUGINS"], "version": version})

    @app.errorhandler(Exception)
    def handle_exception(e):
        if isinstance(e, HTTPException):
            logger.debug(e.description, exc_info=True)
            return jsonify({"error": e.description}), e.code
        if isinstance(e, OperationalError):
            logger.error(e, exc_info=True)
            return {"error": "Database unavailable"}, 504
        logger.error(str(e), exc_info=True)
        return jsonify({"error": "Internal server error occurred"}), 500

    return app
