import os
from flask_migrate import Migrate, Config
from flask_sqlalchemy import SQLAlchemy
from importlib_metadata import EntryPoint
from sqlalchemy import MetaData

from flaschengeist import logger

# https://alembic.sqlalchemy.org/en/latest/naming.html
metadata = MetaData(
    naming_convention={
        "pk": "pk_%(table_name)s",
        "ix": "ix_%(table_name)s_%(column_0_name)s",
        "uq": "uq_%(table_name)s_%(column_0_name)s",
        "ck": "ck_%(table_name)s_%(constraint_name)s",
        "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    }
)


db = SQLAlchemy(metadata=metadata)
migrate = Migrate()


@migrate.configure
def configure_alembic(config: Config):
    """Alembic configuration hook

    Inject all migrations paths into the ``version_locations`` config option.
    This includes even disabled plugins, as simply disabling a plugin without
    uninstall can break the alembic version management.
    """
    from importlib_metadata import entry_points, distribution

    # Set main script location
    config.set_main_option(
        "script_location", str(distribution("flaschengeist").locate_file("") / "flaschengeist" / "alembic")
    )

    # Set Flaschengeist's migrations
    migrations = [config.get_main_option("script_location") + "/migrations"]

    # Gather all migration paths
    ep: EntryPoint
    for ep in entry_points(group="flaschengeist.plugins"):
        try:
            directory = ep.dist.locate_file("")
            for loc in ep.module.split(".") + ["migrations"]:
                directory /= loc
            if directory.exists():
                logger.debug(f"Adding migration version path {directory}")
                migrations.append(str(directory.resolve()))
        except:
            logger.warning(f"Could not load migrations of plugin {ep.name} for database migration.")
            logger.debug("Plugin loading failed", exc_info=True)

    # write back seperator (we changed it if neither seperator nor locations were specified)
    config.set_main_option("version_path_separator", os.pathsep)
    config.set_main_option("version_locations", os.pathsep.join(set(migrations)))
    return config


def case_sensitive(s):
    """
        Compare string as case sensitive on the database

    Args:
        s: string to compare

    Example:
        User.query.filter(User.name == case_sensitive(some_string))
    """
    if db.session.bind.dialect.name == "mysql":
        from sqlalchemy import func

        return func.binary(s)
    return s
